@@ -1,7 +1,9 @@
-from typing import List, Dict, Optional, Union, Tuple, Any, Callable
+from typing import List, Dict, Optional, Union, Tuple, Any, Callable, cast
 import os
 import json
 import math
+import time
+import random
 import requests
 from pathlib import Path
 import asyncio
@@ -16,6 +18,7 @@ from common_pyutil.monitor import Timer
 
 from .filters import (year_filter, author_filter, num_citing_filter,
                       num_influential_count_filter, venue_filter, title_filter)
+from .data import CitationsCache
 
 
 timer = Timer()
@@ -36,13 +39,7 @@ class ConfigType(TypedDict):
     author_papers: SubConfigType
 
 
-CitationType = List[Dict[str, Union[str, Dict[str, str]]]]
-
-
-class CitationsType(TypedDict, total=False):
-    next: int
-    offset: int
-    data: CitationType
+ContextType = str
 
 
 class DetailsDataType(TypedDict):
@@ -53,8 +50,20 @@ class DetailsDataType(TypedDict):
     venue: str
     year: str
     authors: List[Dict]
-    citations: CitationsType
-    references: CitationsType
+    # citations: "CitationsType"
+    # references: "CitationsType"
+    externalIds: Dict[str, Union[int, str]]
+
+
+class CitationType(TypedDict):
+    contexts: List[str]
+    citingPaper: DetailsDataType
+
+
+class CitationsType(TypedDict, total=False):
+    next: int
+    offset: int
+    data: List[CitationType]
 
 
 class StoredDataType(TypedDict):
@@ -66,15 +75,33 @@ class StoredDataType(TypedDict):
 Cache = Dict[str, Dict[str, str]]
 
 
+def get_corpus_id(data: Union[CitationType, DetailsDataType]) -> int:
+    if "externalIds" in data:
+        if data["externalIds"]:                   # type: ignore
+            cid = data["externalIds"]["CorpusId"]  # type: ignore
+            return int(cid)
+    elif "citingPaper" in data:
+        if data["citingPaper"]["externalIds"]:                   # type: ignore
+            cid = data["citingPaper"]["externalIds"]["CorpusId"]  # type: ignore
+            return int(cid)
+    return -1
+
+
+def citations_corpus_ids(data: Dict) -> List[int]:
+    return [int(x["citingPaper"]["externalIds"]["CorpusId"])
+            for x in data["citations"]["data"]]
+
+
+
 class SemanticScholar:
-    """A Semantic Scholar API client with a files based Cache.
+    """A Semantic Scholar API client with a files based cache.
 
     The cache is a Dictionary of type :code:`Cache` where they keys are one of
     `["acl", "arxiv", "corpus", "doi", "mag", "url"]` and values are a dictionary
-    of that id type and the associated ss_id.
+    of that id type and the associated :code:`ss_id`.
 
-    Each ss_id is stored as a file with the same
-    name as the ss_id and contains the data for the entry in JSON format.
+    Each :code:`ss_id` is stored as a file with the same
+    name as the :code:`ss_id` and contains the data for the entry in JSON format.
 
     Args:
         root: root directory where all the metadata and the
@@ -99,13 +126,14 @@ class SemanticScholar:
                                          "title": title_filter}
         return _filters
 
-    def __init__(self, config_file=None, cache_dir=None):
+    def __init__(self, config_file=None, cache_dir=None, refs_cache_dir=None):
         self.load_config(config_file)
         self._api_key = self._config.get("api_key", None)
         self._root_url = "https://api.semanticscholar.org/graph/v1"
         self._cache_dir = Path(cache_dir or self._config.get("cache", None))
         if not self._cache_dir or not Path(self._cache_dir).exists():
             raise FileExistsError(f"{self._cache_dir} doesn't exist")
+        self._refs_cache_dir = refs_cache_dir
         self._in_memory: Dict[str, StoredDataType] = {}
         self._cache: Cache = {}
         self._rev_cache: Dict[str, List[str]] = {}
@@ -113,7 +141,22 @@ class SemanticScholar:
                                           os.listdir(self._cache_dir))]
         self._id_keys = ['DOI', 'MAG', 'ARXIV', 'ACL', 'PUBMED', 'URL', 'CorpusId']
         self._id_keys.sort()
+        # NOTE: Actually the SS limit is 100 reqs/sec. This applies to all requests I think
+        #       and there's no way to change this. This is going to fetch fairly slowly
+        #       Since my timeout is 5 secs, I can try fetching with 500
+        self._batch_size = 500
+        self._tolerance = 2
+        self._dont_build_citations = set()
+        self._aio_timeout = aiohttp.ClientTimeout(10)
         self.load_metadata()
+        self.load_refs_cache()
+
+    def load_refs_cache(self):
+        if self._refs_cache_dir and Path(self._refs_cache_dir).exists():
+            self._refs_cache = CitationsCache(self._refs_cache_dir)
+            print(f"Loaded Full Semantic Scholar Citations Cache from {self._refs_cache_dir}")
+        else:
+            self._refs_cache = None
 
     def id_types(self, id_type: str):
         return "CorpusId" if id_type.lower() == "corpusid" else id_type.upper()
@@ -183,6 +226,13 @@ class SemanticScholar:
                                              'influentialCitationCount',
                                              'externalIds']}}
 
+    @property
+    def headers(self) -> Dict[str, str]:
+        if self._api_key:
+            return {"x-api-key": self._api_key}
+        else:
+            return {}
+
     def load_metadata(self):
         """Load the Semantic Scholar metadata from the disk.
 
@@ -256,10 +306,7 @@ class SemanticScholar:
             url: URL
 
         """
-        if self._api_key:
-            response = requests.get(url, headers={"x-api-key": self._api_key})
-        else:
-            response = requests.get(url)
+        response = requests.get(url, headers=self.headers)
         return response
 
     def _dump(self, ID: str, data: StoredDataType):
@@ -271,15 +318,18 @@ class SemanticScholar:
                 json.dump(data, f)
         print(f"Wrote file {fname} in {timer.time} seconds")
 
-    def _update_citations(self, data: CitationsType,
-                          existing_data: CitationsType):
+    def _update_citations(self, data_a: CitationsType,
+                          data_b: CitationsType):
+        """Update from data_a from data_b"""
         # NOTE: Ignoring as the Dict isn't uniform and it raises error
-        data_cite_ids = {x["citingPaper"]["paperId"] for x in data["data"]}  # type: ignore
-        for x in existing_data["data"]:
-            if x["citingPaper"]["paperId"] not in data_cite_ids:  # type: ignore
-                data["data"].append(x)
-                if "next" in data:
-                    data["next"] += 1
+        data_a["data"] = [*filter(lambda x: "error" not in x["citingPaper"], data_a["data"])]
+        data_b["data"] = [*filter(lambda x: "error" not in x["citingPaper"], data_b["data"])]
+        data_ids_b = {x["citingPaper"]["paperId"] for x in data_b["data"]}  # type: ignore
+        for x in data_a["data"]:
+            if x["citingPaper"]["paperId"] not in data_ids_b:  # type: ignore
+                data_b["data"].append(x)
+                if "next" in data_b:
+                    data_b["next"] += 1
 
     def put_all(self, data: StoredDataType):
         """Update paper details, references and citations on disk.
@@ -298,7 +348,7 @@ class SemanticScholar:
         # NOTE: In case force updated and already some citations exist on disk
         existing_data = self._check_cache(paper_id)
         if existing_data is not None:
-            self._update_citations(data["citations"], existing_data["citations"])
+            self._update_citations(existing_data["citations"], data["citations"])
         self._dump(paper_id, data)
         ext_ids = {self.id_types(k): v for k, v in details["externalIds"].items()}  # type: ignore
         other_ids = [ext_ids.get(k, "") for k in self._id_keys]  # type: ignore
@@ -378,13 +428,10 @@ class SemanticScholar:
 
     async def _author(self, ID: str) -> Dict:
         urls = [f(ID) for f in [self.author_url, self.author_papers_url]]
-        if self._api_key:
-            headers = {"x-api-key": self._api_key}
-        else:
-            headers = {}
-        async with aiohttp.ClientSession(headers=headers) as session:
-            tasks = [self._aget(session, url) for url in urls]
-            results = await asyncio.gather(*tasks)
+        results = await self._get_some_urls(urls)
+        # async with aiohttp.ClientSession(headers=self.headers) as session:
+        #     tasks = [self._aget(session, url) for url in urls]
+        #     results = await asyncio.gather(*tasks)
         return dict(zip(["author", "papers"], results))
 
     def get_author_papers(self, ID: str) -> Dict:
@@ -404,6 +451,28 @@ class SemanticScholar:
         data = await resp.json()
         return data
 
+    async def _get_some_urls(self, urls: List[str], timeout: Optional[int] = None) -> List:
+        """Get some URLs asynchronously
+
+        Args:
+            urls: List of URLs
+
+        URLs are fetched with :class:`aiohttp.ClientSession` with api_key included
+
+        """
+        if not timeout:
+            timeout = self._aio_timeout
+        else:
+            timeout = aiohttp.ClientTimeout(timeout)  # type: ignore
+        try:
+            async with aiohttp.ClientSession(headers=self.headers,
+                                             timeout=timeout) as session:
+                tasks = [self._aget(session, url) for url in urls]
+                results = await asyncio.gather(*tasks)
+        except asyncio.exceptions.TimeoutError:
+            return []
+        return results
+
     async def _paper(self, ID: str) -> StoredDataType:
         """Asynchronously fetch paper details, references and citations.
 
@@ -416,13 +485,11 @@ class SemanticScholar:
         urls = [f(ID) for f in [self.details_url,  # type: ignore
                                 self.references_url,
                                 self.citations_url]]
-        if self._api_key:
-            headers = {"x-api-key": self._api_key}
-        else:
-            headers = {}
-        async with aiohttp.ClientSession(headers=headers) as session:
-            tasks = [self._aget(session, url) for url in urls]
-            results = await asyncio.gather(*tasks)
+        print("FETCHING paper with _get_some_urls")
+        results = await self._get_some_urls(urls)
+        # async with aiohttp.ClientSession(headers=self.headers) as session:
+        #     tasks = [self._aget(session, url) for url in urls]
+        #     results = await asyncio.gather(*tasks)
         # NOTE: mypy can't resolve zip of async gather
         data: StoredDataType = dict(zip(["details", "references", "citations"], results))  # type: ignore
         return data
@@ -475,7 +542,29 @@ class SemanticScholar:
             print(f"Fetching from Semantic Scholar for {ID}")
             return self.store_details_and_get(ID, no_transform)
 
-    def get_details_for_id(self, id_type: str, ID: str, force: bool) -> Union[str, Dict]:
+    # TODO: Don't we expect the other data to be in cache?
+    def _corpus_id(self, id_type: str, ID: str) -> Union[str, int]:
+        ids = {"doi": f"DOI:{ID}",
+               "mag": f"MAG:{ID}",
+               "arxiv": f"ARXIV:{ID}",
+               "acl": f"ACL:{ID}",
+               "pubmed": f"PMID:{ID}",
+               "url": f"URL:{ID}"}
+        if id_type == "ss":
+            ssid = ID
+            have_metadata = ssid in self._rev_cache
+        elif id_type not in ids:
+            return "INVALID ID TYPE"
+        else:
+            itype = self.id_types(id_type)
+            ssid = self._cache[itype].get(ID, "")
+            have_metadata = bool(ssid)
+        data = self.fetch_from_cache_or_service(
+            have_metadata, ssid or ids[id_type], False, False)
+        return data['externalIds']['CorpusId']
+
+    def get_details_for_id(self, id_type: str, ID: str, force: bool, all_data: bool)\
+            -> Union[str, Dict]:
         """Get paper details from Semantic Scholar Graph API
 
         The on disk cache is checked first and if it's a miss then the
@@ -489,7 +578,8 @@ class SemanticScholar:
                      `['ss', 'doi', 'mag', 'arxiv', 'acl', 'pubmed', 'corpus']`
             ID: paper identifier
             force: Force fetch from Semantic Scholar server, ignoring cache
-
+            all_data: Fetch all details if possible. This can only be done if
+                      the data already exists on disk
         """
         ids = {"doi": f"DOI:{ID}",
                "mag": f"MAG:{ID}",
@@ -507,12 +597,16 @@ class SemanticScholar:
             itype = self.id_types(id_type)
             ssid = self._cache[itype].get(ID, "")
             have_metadata = bool(ssid)
-        return self.apply_limits(self.fetch_from_cache_or_service(
-            have_metadata, ssid or ids[id_type], force, False))
+        data = self.fetch_from_cache_or_service(
+            have_metadata, ssid or ids[id_type], force, False)
+        if all_data:
+            return data
+        else:
+            return self.apply_limits(data)
 
     # CHECK: Is this function here just because I have references and citations?
     #        For consistency?
-    def details(self, ID: str, force: bool = False) -> Union[str, Dict]:
+    def details(self, ID: str, force: bool = False, all_data: bool = False) -> Union[str, Dict]:
         """Get details for paper with SSID :code:`ID`
 
         Args:
@@ -520,7 +614,7 @@ class SemanticScholar:
             force: Whether to force fetch from service
 
         """
-        return self.get_details_for_id("ss", ID, force)
+        return self.get_details_for_id("ss", ID, force, all_data)
 
     def apply_limits(self, data):
         """Apply count limits to S2 data citations and references
@@ -540,23 +634,13 @@ class SemanticScholar:
             _data["references"] = _data["references"][:limit]
         return _data
 
-    # TODO: Do I need two details functions?
-    #       See details above
-    # def get_all_details(self, ID: str):
-    #     """Get details for paper with SSID :code:`ID`
-
-    #     Args:
-    #         ID: SSID of the paper
-
-    #     Difference between this and :meth:`details` is that this one calls
-    #     :meth:`fetch_from_cache_or_service` directly.
-
-    #     """
-    #     ssid = ID
-    #     have_metadata = ssid in self._rev_cache
-    #     return self.fetch_from_cache_or_service(have_metadata, ssid, False, True)
-
     def _get_details_from_disk(self, ID: str) -> Optional[StoredDataType]:
+        """Fetch S2 details from disk with SSID=ID
+
+        Args:
+            ID: SSID of the paper
+
+        """
         data_file = self._cache_dir.joinpath(ID)
         if data_file.exists():
             print(f"Data for {ID} is on disk")
@@ -611,6 +695,10 @@ class SemanticScholar:
             data = self._get_details_from_disk(ID)
             if data:
                 if self._validate_fields(data):
+                    data["citations"]["data"] =\
+                        [x for x in data["citations"]["data"] if "paperId" in x["citingPaper"]]
+                    data["references"]["data"] =\
+                        [x for x in data["references"]["data"] if "paperId" in x["citedPaper"]]
                     self._in_memory[ID] = data
                 else:
                     print(f"Stale data for {ID}")
@@ -654,18 +742,22 @@ class SemanticScholar:
 
         """
         data = self.fetch_from_cache_or_service(True, ID, False, True)
+        data = cast(StoredDataType, data)
         limit = count or self._config["citations"]["limit"]
         cite_data = data["citations"]["data"]
-        if offset:
-            if offset + limit > len(cite_data):
-                self.next_citations(ID, limit, offset)
-                data = self._check_cache(ID)  # type: ignore
-                cite_data = data["citations"]["data"]
-            retval = cite_data[offset:offset+limit]
-        else:
-            retval = cite_data[:limit]
+        total_cites = data["details"]["citationCount"]
+        if offset + limit > total_cites:
+            limit = total_cites - offset
+        if offset + limit > len(cite_data):
+            self.next_citations(ID, limit, offset)
+            data = self._check_cache(ID)  # type: ignore
+            cite_data = data["citations"]["data"]
+        retval = cite_data[offset:offset+limit]
         return [x["citingPaper"] for x in retval]
 
+    # TODO: Although this fetches the appropriate data based on citations on disk
+    #       the offset and limit handling is tricky and is not correct right now.
+    # TODO: What if the num_citations change between the time we fetched earlier and now?
     def next_citations(self, ID: str, limit: int = 0, offset: int = 0) -> Optional[Dict]:
         """Fetch next citations for a paper if any.
 
@@ -675,6 +767,10 @@ class SemanticScholar:
         Args:
             ID: The paper ID
 
+        There is an issue with S2 API in that even though it may show :code:`n`
+        number of citing papers, when actually fetching the citation data it
+        retrieves fewer citations sometimes.
+
         """
         limit = limit or self._config["citations"]["limit"]
         data = self._check_cache(ID)
@@ -683,91 +779,221 @@ class SemanticScholar:
         elif data is not None and "next" not in data["citations"]:
             return None
         else:
-            paper_id = data["details"]["paperId"]
-            _next = data["citations"]["next"]
-            if offset:
-                limit += offset - _next
-            offset = _next
-            url = self.citations_url(paper_id, limit, offset)
-            citations = json.loads(self._get(url).content)
-            self._update_citations(data["citations"], citations)
+            cite_count = data["details"]["citationCount"]
+            if offset+limit > 10000:
+                corpus_id = get_corpus_id(data["details"])
+                if corpus_id:
+                    citations = self._build_citations_from_stored_data(
+                        corpus_id, citations_corpus_ids(data),
+                        cite_count, offset, limit)
+            else:
+                paper_id = data["details"]["paperId"]
+                _next = data["citations"]["next"]
+                if offset:
+                    limit += offset - _next
+                offset = _next
+                url = self.citations_url(paper_id, limit, offset)
+                citations = json.loads(self._get(url).content)
+            self._update_citations(citations, data["citations"])
             self._dump(paper_id, data)
             return citations
 
-    def filter_subr(self, key: str, values: CitationType, filters: Dict[str, Any],
+    def filter_subr(self, key: str, citation_data: List[CitationType], filters: Dict[str, Any],
                     num: int) -> List[Dict]:
-        """Subroutine for filtering papers
+        """Subroutine for filtering references and citations
 
         Each filter function is called with the arguments and the results are AND'ed.
 
         Args:
             key: One of "references" or "citations"
-            values: The values (paper details) to filter
+            citation_data: citation (or references) data to filter
             filters: Filter names and kwargs
 
         """
         retvals = []
-        for val in values:
+        for citation in citation_data:
             status = True
-            for k, v in filters.items():
-                if key in val:
+            for filter_name, filter_args in filters.items():
+                # key is either citedPaper or citingPaper
+                # This is a bit redundant as key should always be there but this will
+                # catch edge cases
+                if key in citation:
                     try:
                         # kwargs only
-                        status = status and self.filters[k](val[key], **v)
+                        filter_func = self.filters[filter_name]
+                        status = status and filter_func(citation[key], **filter_args)
                     except Exception as e:
-                        print(f"Can't apply filter {k} on {val}: {e}")
+                        print(f"Can't apply filter {filter_name} on {citation}: {e}")
                         status = False
                 else:
                     status = False
             if status:
-                retvals.append(val)
+                retvals.append(citation)
             if num and len(retvals) == num:
                 break
         # NOTE: Gives error because x[key] evals to Union[str, Dict[str, str]]
         return [x[key] for x in retvals]
 
-    async def _ensure_all_citations(self, ID: str) -> CitationsType:
+    # TODO: This should either use `next_citations` or `next_citations` should use this
+    # FIXME: This function is too long
+    def _ensure_all_citations(self, ID: str) -> CitationsType:
         data = self._check_cache(ID)
         if data is not None:
-            num_cites = data["details"]["citationCount"]
-            offset = data["citations"].get("next", None)
-            if offset:
-                fields = ",".join(self._config["citations"]["fields"])
-                iters = math.ceil((num_cites - offset) / 1000)
-                urls = [f"{self._root_url}/paper/{ID}/citations?fields={fields}&limit=1000"
-                        + f"&offset={offset + (i*1000)}"
-                        for i in range(iters)]
-                if self._api_key:
-                    headers = {"x-api-key": self._api_key}
+            cite_count = data["details"]["citationCount"]
+            existing_cite_count = len(data["citations"]["data"])
+            # NOTE: We should always instead get from the head of the stream
+            #       and then merge
+            offset = 0
+            if cite_count > 10000:
+                print("Warning: More than 10000 citations cannot be fetched. "
+                      "Will only get first 10000")
+            fields = ",".join(self._config["citations"]["fields"])
+            iters = min(10, math.ceil((cite_count - existing_cite_count) / 1000))
+            urls = []
+            for i in range(iters):
+                # limit = min(1000, 10000 - (offset + i * 1000))
+                limit = 9999 - (offset + i * 1000)\
+                    if (offset + i * 1000) + 1000 > 10000 else 1000
+                urls.append(f"{self._root_url}/paper/{ID}/citations?"
+                            f"fields={fields}&limit={limit}"
+                            f"&offset={offset + i * 1000}")
+            print(f"Will fetch {len(urls)} requests for citations")
+            print(f"All urls {urls}")
+            with timer:
+                results = asyncio.run(self._get_some_urls(urls))
+            print(f"Got {len(results)} results")
+            result: CitationsType = {"next": 0, "offset": 0, "data": []}
+            cite_list = []
+            errors = 0
+            for x in results:
+                if "error" not in x:
+                    cite_list.extend(x["data"])
                 else:
-                    headers = {}
-                print(f"Will fetch {len(urls)} requests for citations")
-                async with aiohttp.ClientSession(headers=headers) as session:
-                    tasks = [self._aget(session, url) for url in urls]
-                    results = await asyncio.gather(*tasks)
-                result: CitationsType = {"next": 0, "offset": 0, "data": []}
-                cite_list = []
-                errors = 0
-                for x in results:
-                    if "error" not in x:
-                        cite_list.extend(x["data"])
-                    else:
-                        errors += 1
-                result["data"] = cite_list
-                if errors:
-                    print(f"{errors} errors occured while fetching all citations for {ID}")
-                offset = max(x["offset"] for x in results if "error" not in x)
-                if all("next" in x for x in results):
-                    result["next"] = max(x["next"] for x in results if "error" not in x)
-                else:
-                    result.pop("next")
-                return result
+                    errors += 1
+            result["data"] = cite_list
+            print(f"Have {len(cite_list)} citations without errors")
+            if errors:
+                print(f"{errors} errors occured while fetching all citations for {ID}")
+            offset = max([*[x["offset"] for x in results if "error" not in x], 10000])
+            if all("next" in x for x in results):
+                result["next"] = max([*[x["next"] for x in results if "error" not in x], 10000])
             else:
-                return data["citations"]
+                result.pop("next")
+            return result
         else:
             msg = f"Paper data for {ID} should already exist"
             raise ValueError(msg)
 
+    def _get_some_urls_in_batches(self, urls: List[str], batch_size: int) -> List[Dict]:
+        """Fetch batch_size examples at a time to prevent overloading the service
+
+        Args:
+            urls: URLs to fetch
+            batch_size: batch_size
+
+        """
+        j = 0
+        _urls = urls[j*batch_size:(j+1)*batch_size]
+        results = []
+        _results = []
+        while _urls:
+            print(f"Fetching for j {j} out of {len(urls)//batch_size} urls")
+            with timer:
+                _results = asyncio.run(self._get_some_urls(_urls, 5))
+                while not _results:
+                    wait_time = random.randint(1, 5)
+                    print(f"Got empty results. Waiting {wait_time}")
+                    time.sleep(wait_time)
+                    _results = asyncio.run(self._get_some_urls(_urls))
+            results.extend(_results)
+            j += 1
+            _urls = urls[j*batch_size:(j+1)*batch_size]
+        return results
+
+    # TODO: Need to add condition such that if num_citations > 10000, then this
+    #       function is called. And also perhaps, fetch first 1000 citations and
+    #       update the stored data (if they're sorted by time)
+    def _build_citations_from_stored_data(self,
+                                          corpus_id: Union[int, str],
+                                          existing_ids: Optional[List[int]],
+                                          cite_count: int,
+                                          offset: int = 0,
+                                          limit: int = 0) -> CitationsType:
+        """Build the citations data for a paper entry from cached data
+
+        Args:
+            corpus_id: Semantic Scholar CorpusId
+
+        """
+
+        refs_ids = self._refs_cache.get_citations(int(corpus_id))
+        if not refs_ids:
+            raise AttributeError(f"Not found for {corpus_id}")
+        if not existing_ids:
+            existing_ids = []
+        if existing_ids:
+            need_ids = list(refs_ids - set(existing_ids))
+        else:
+            need_ids = list(refs_ids)
+        if not limit:
+            limit = len(need_ids)
+        cite_gap = cite_count - len(need_ids) - len(existing_ids)
+        if cite_gap:
+            print(f"WARNING: {cite_gap} citations cannot be fetched.\n"
+                  "You have stale SS data")
+        need_ids = need_ids[offset:offset+limit]
+        # Remove contexts as that's not available in paper details
+        fields = ",".join(self._config["citations"]["fields"]).replace(",contexts", "")
+        urls = [f"{self._root_url}/paper/CorpusID:{ID}?fields={fields}"
+                for ID in need_ids]
+        citations = {"offset": 0, "data": []}
+        batch_size = self._batch_size
+        result = self._get_some_urls_in_batches(urls, batch_size)
+        citations["data"] = [{"citingPaper": x, "contexts": []} for x in result]
+        return citations        # type: ignore
+
+    def _fetch_citations_greater_than_10000(self, existing_data):
+        """Fetch citations when their number is > 10000.
+
+        SemanticScholar doesn't allow above 10000, so we have to build that
+        from the dumped citation data. See :meth:`_build_citations_from_stored_data`
+
+        Args:
+            existing_data: Existing paper details data
+
+        """
+        cite_count = len(existing_data["citations"]["data"])
+        existing_corpus_ids = [get_corpus_id(x) for x in existing_data["citations"]["data"]]
+        if -1 in existing_corpus_ids:
+            existing_corpus_ids.remove(-1)
+        corpus_id = get_corpus_id(existing_data["details"])  # type: ignore
+        if not corpus_id:
+            raise AttributeError("Did not expect corpus_id to be 0")
+        if corpus_id not in self._dont_build_citations:
+            more_data = self._build_citations_from_stored_data(corpus_id,
+                                                               existing_corpus_ids,
+                                                               cite_count)
+            new_ids = set([x["citingPaper"]["paperId"]
+                           for x in more_data["data"]
+                           if "citingPaper" in x and "error" not in x["citingPaper"]
+                           and "paperId" in x["citingPaper"]])
+            # NOTE: Some debug vars commented out
+            # new_data_dict = {x["citingPaper"]["paperId"]: x["citingPaper"]
+            #                  for x in more_data["data"]
+            #                  if "citingPaper" in x and "error" not in x["citingPaper"]
+            #                  and "paperId" in x["citingPaper"]}
+            # existing_citation_dict = {x["citingPaper"]["paperId"]: x["citingPaper"]
+            #                           for x in existing_data["citations"]["data"]}
+            existing_ids = set(x["citingPaper"]["paperId"]
+                               for x in existing_data["citations"]["data"])
+            something_new = new_ids - existing_ids
+            if more_data["data"] and something_new:
+                print(f"Fetched {len(more_data['data'])} in {timer.time} seconds")
+                self._update_citations(more_data, existing_data["citations"])
+                update = True
+        self._dont_build_citations.add(corpus_id)
+        return update
+
     def filter_citations(self, ID: str, filters: Dict[str, Any], num: int = 0) -> List[Dict]:
         """Filter citations based on given filters.
 
@@ -787,18 +1013,27 @@ class SemanticScholar:
             msg = f"data should not be None for ID {ID}"
             return msg          # type: ignore
         else:
-            if not existing_data["details"]["citationCount"] ==\
-               len(existing_data["citations"]["data"]):
+            update = False
+            cite_count = existing_data["details"]["citationCount"]
+            existing_cite_count = len(existing_data["citations"]["data"])
+            if abs(cite_count - existing_cite_count) > self._tolerance:
                 with timer:
-                    data = asyncio.run(self._ensure_all_citations(ID))
+                    data = self._ensure_all_citations(ID)
                 print(f"Fetched {len(data['data'])} in {timer.time} seconds")
                 self._update_citations(data, existing_data["citations"])
-                existing_data["citations"] = data
-                self._dump(ID, existing_data)
+                if len(existing_data["citations"]["data"]) > existing_cite_count:
+                    print(f"Fetched {len(existing_data['citations']['data']) - existing_cite_count}"
+                          " new citations")
+                    update = True
+                if cite_count > 10000:
+                    _update = self._fetch_citations_greater_than_10000(existing_data)
+                    update = update or _update
+                if update:
+                    self._dump(ID, existing_data)
             return self.filter_subr("citingPaper", existing_data["citations"]["data"], filters, num)
 
     def filter_references(self, ID: str, filters: Dict[str, Any], num: int = 0):
-        """LIke :meth:`filter_citations` but for references
+        """Like :meth:`filter_citations` but for references
 
 
         """
@@ -825,5 +1060,3 @@ class SemanticScholar:
             return response.content
         else:
             return json.dumps({"error": json.loads(response.content)})
-
-
