@@ -47,10 +47,11 @@ class RefMan:
                                Used by supported methods.
         data_dir: Directory where the Semantic Scholar Cache is stored.
                   See :func:`load_ss_cache`
+        config_dir: Directory to store configuration related files
+        refs_cache_dir: Directory of Semantic Scholar References Cache
         local_pdfs_dir: Local directory where the pdf files are stored.
         remote_pdfs_dir: Remote directory where the pdf files are stored.
         remote_links_cache: File mapping local pdfs to remote links.
-        config_dir: Directory to store configuration related files
         batch_size: Number of parallel requests to send in case parallel requests is
                     implemented for that method.
         chrome_debugger_path: Path for the chrome debugger script.
@@ -67,17 +68,21 @@ class RefMan:
 
     """
     # TODO: Generate default config and place in config dir.
-    #       Also allow the user to add config options like request headers,
+    #       As of now only SemanticScholar configuration is loaded if it exists
+    # TODO: Also allow the user to add config options like request headers,
     #       proxy ports, data_dir etc.
     def __init__(self, host: str, port: int, proxy_port: int, proxy_everything: bool,
-                 proxy_everything_port: int, data_dir: Path, local_pdfs_dir: Path,
-                 remote_pdfs_dir: str, remote_links_cache: Path,
-                 config_dir: Path, batch_size: int, chrome_debugger_path: str,
+                 proxy_everything_port: int, data_dir: Path, refs_cache_dir: Path,
+                 config_dir: Path, local_pdfs_dir: Path,
+                 remote_pdfs_dir: Path, remote_links_cache: Path,
+                 batch_size: int, chrome_debugger_path: Path,
                  debug: bool, verbosity: str, threaded: bool):
         self.host = "127.0.0.1"
         self.port = port
         self.batch_size = batch_size
-        self.data_dir = Path(data_dir)
+        self.data_dir = data_dir
+        self.config_dir = config_dir
+        self.refs_cache_dir = refs_cache_dir
         self.local_pdfs_dir = local_pdfs_dir
         self.remote_pdfs_dir = remote_pdfs_dir
         self.remote_links_cache = remote_links_cache
@@ -85,7 +90,6 @@ class RefMan:
         self.proxy_everything = proxy_everything
         self.proxy_everything_port = proxy_everything_port
         self.chrome_debugger_path = Path(chrome_debugger_path) if chrome_debugger_path else None
-        self.config_dir = config_dir
         if not self.config_dir.exists():
             os.makedirs(self.config_dir)
         self.config_file: Optional[Path] = self.config_dir.joinpath("config.json")
@@ -94,9 +98,13 @@ class RefMan:
         self.verbosity = verbosity
         self.threaded = threaded
 
-        self.set_verbosity()
+        self.init_loggers()
+        if self.config_file:
+            self.logi(f"Loaded config file {self.config_file}")
+
         self.load_cvf_files()
-        self.s2 = SemanticScholar(cache_dir=self.data_dir, config_file=self.config_file)
+        self.s2 = SemanticScholar(config_file=self.config_file, cache_dir=self.data_dir,
+                                  refs_cache_dir=self.refs_cache_dir)
         self.init_remote_cache()
 
         # NOTE: Checks only once for the proxy, see util.check_proxy
@@ -124,7 +132,7 @@ class RefMan:
             self.logger.warning("All arguments required for pdf cache not given.\n" +
                                 "Will not maintain remote pdf links cache.")
 
-    def set_verbosity(self):
+    def init_loggers(self):
         # We set "error" to warning
         verbosity_levels = {"info", "error", "debug"}
         if self.verbosity not in verbosity_levels:
@@ -147,8 +155,8 @@ class RefMan:
                           for f in os.listdir(self.config_dir)
                           if re.match(r'^(cvpr|iccv)', f.lower())]
         self.soups = {}
-        self.cvf_url_root = "https://openaccess.thecvf.com/"
-        self.logger.debug(f"Loading CVF soups.")
+        self.cvf_url_root = "https://openaccess.thecvf.com"
+        self.logger.debug("Loading CVF soups.")
         for cvf in self.cvf_files:
             match = re.match(r'^(cvpr|iccv)(.*?)([0-9]+)',
                              Path(cvf).name, flags=re.IGNORECASE)
@@ -160,13 +168,12 @@ class RefMan:
                 self.logger.error(f"Could not load file {cvf}")
         self.logger.debug(f"Loaded conference files {self.soups.keys()}")
 
-    def download_cvf_page_and_update_soups(self, venue, year):
-        url = f"{self.cvf_url_root}/{venue.upper()}{year}"
-        resp = requests.get(url)
-        soup = BeautifulSoup(resp.content, features="lxml")
+    def maybe_download_cvf_day_pages(self, response, venue, year):
+        soup = BeautifulSoup(response.content, features="lxml")
         links = soup.find_all("a")
         regexp = f"{venue.upper()}{year}.py"
-        if re.match(regexp + ".+", links[-1].attrs['href']):
+        last_link_attrs = links[-1].attrs
+        if "href" in last_link_attrs and re.match(regexp + ".+", last_link_attrs['href']):
             day_links = [*filter(lambda x: re.match(r"Day [0-9]+?: ([0-9-+])", x.text),
                                  soup.find_all("a"))]
             content = []
@@ -174,18 +181,28 @@ class RefMan:
                 day = re.match(r"Day [0-9]+?: ([0-9-]+)", dl.text).groups()[0]
                 d_url = f"{self.cvf_url_root}/{venue.upper()}{year}.py?day={day}"
                 resp = requests.get(d_url)
+                if resp.status_code != 200:
+                    raise requests.HTTPError(f"Status code {response.status_code} for {d_url}")
                 content.append(resp.content)
                 self.logd(f"Fetched page {i+1} for {venue.upper()}{year} and {day}")
             content = "\n".join([x.decode() for x in content])
         else:
             self.logd(f"Fetched page for {venue.upper()}{year}")
-            content = resp.content.decode()
+            content = response.content.decode()
+        return content
+
+    def download_cvf_page_and_update_soups(self, venue, year):
+        url = f"{self.cvf_url_root}/{venue.upper()}{year}"
+        response = requests.get(url)
+        if response.status_code == 200:
+            content = self.maybe_download_cvf_day_pages(response, venue, year)
+        else:
+            raise requests.HTTPError(f"Status code {response.status_code} for {url}")
         fname = self.config_dir.joinpath(f"{venue.upper()}{year}")
-        if resp.status_code == 200:
-            with open(fname, "w") as f:
-                f.write(content)
+        with open(fname, "w") as f:
+            f.write(content)
         with open(fname) as f:
-            self.soups[(venue.lower(), year)] = BeautifulSoup(f.read(), features="lxml")
+            self.soups[(venue.lower(), year)] = BeautifulSoup(content, features="lxml")
 
     def logi(self, msg: str) -> str:
         self.logger.info(msg)
@@ -261,15 +278,26 @@ class RefMan:
                     id_type = request.args["id_type"]
                 else:
                     return json.dumps("NO ID_TYPE GIVEN")
-                if "force" in request.args:
-                    force = True
-                else:
-                    force = False
-                data = self.s2.get_details_for_id(id_type, id, force)
+                force = True if "force" in request.args else False
+                all_data = True if "all" in request.args else False
+                data = self.s2.get_details_for_id(id_type, id, force, all_data)
                 return json.dumps(data)
             else:
                 return json.dumps("METHOD NOT IMPLEMENTED")
 
+        @app.route("/s2_corpus_id", methods=["GET"])
+        def s2_corpus_id():
+            if "id" in request.args:
+                id = request.args["id"]
+            else:
+                return json.dumps("NO ID GIVEN")
+            if "id_type" in request.args:
+                id_type = request.args["id_type"]
+            else:
+                return json.dumps("NO ID_TYPE GIVEN")
+            data = self.s2.get_corpus_id(id_type, id)
+            return json.dumps(data)
+
         @app.route("/s2_config", methods=["GET"])
         def s2_config():
             return json.dumps(self.s2._config)
@@ -282,13 +310,15 @@ class RefMan:
                 force = False
             return json.dumps(self.s2.details(ssid, force))
 
-        # @app.route("/s2_all_details/<ssid>", methods=["GET", "POST"])
-        # def s2_all_details(ssid: str) -> Union[str, bytes]:
-        #     """Get paper, metadata, references and citations for SSID."""
-        #     if request.method == "GET":
-        #         return json.dumps(self.s2.get_all_details(ssid))
-        #     else:
-        #         return json.dumps("METHOD NOT IMPLEMENTED")
+        @app.route("/s2_all_details/<ssid>", methods=["GET", "POST"])
+        def s2_all_details(ssid: str) -> Union[str, bytes]:
+            """Get paper, metadata, references and ALL citations for SSID."""
+            if "force" in request.args:
+                print("Force not supported with s2_all_details")
+            if request.method == "GET":
+                return json.dumps(self.s2.details(ssid, all_data=True))
+            else:
+                return json.dumps("METHOD NOT IMPLEMENTED")
 
         def s2_citations_references_subr(request, ssid: str, key) -> Union[str, bytes]:
             """Get requested citations or references for a paper.
@@ -438,19 +468,39 @@ class RefMan:
             else:
                 return json.dumps("NO URL or URLs GIVEN")
 
-        @app.route("/fetch_proxy")
-        def fetch_proxy():
-            """Fetch URL with :attr:`self.proxies` if :attr:`self.proxies` is not `None`.
+        @app.route("/set_proxy")
+        def set_proxy():
+            """Set or unset proxy port"""
+            if request.args.get("set"):
+                if request.args.get("proxy_port"):
+                    self.proxy_port = request.args.get("proxy_port")
+                if request.args.get("proxy_everything_port"):
+                    self.proxy_everything_port = request.args.get("proxy_everything_port")
+                return self.check_proxies()
+            elif request.args.get("unset"):
+                self.proxy_port = None
+                self.proxy_everything_port = None
+                self.proxies = None
+                return "Unset all proxies"
+
+        @app.route("/fetch_url")
+        def fetch_url():
+            """Fetch given URL.
+
+            Optionally if :attr:`self.proxies` if :attr:`self.proxies` is
+            not :code:`None` then fetch via those proxies
+
             """
             if "url" in request.args and request.args["url"]:
                 url = request.args["url"]
             else:
                 return json.dumps("NO URL GIVEN or BAD URL")
+            noproxy = request.args.get("noproxy")
             keys = [*request.args.keys()]
             # NOTE: Rest of the keys are part of the URL
             if len(keys) > 1:
                 url = url + "&" + "&".join([f"{k}={v}" for k, v in request.args.items()
-                                            if k != "url"])
+                                            if k not in {"url", "noproxy"}])
             # DEBUG code
             # if url == "https://arxiv.org/pdf/2006.01912":
             #     with os.path.expanduser("~/pdf_file.pdf", "rb") as f:
@@ -459,7 +509,7 @@ class RefMan:
             #     response.headers["Content-Type"] = "application/pdf"
             #     return response
             self.logger.debug(f"Fetching {url} with proxies {self.proxies}")
-            if self.proxies:
+            if not noproxy and self.proxies:
                 try:
                     response = requests.get(url, headers=default_headers, proxies=self.proxies)
                 except requests.exceptions.Timeout:
@@ -471,7 +521,8 @@ class RefMan:
                     self.proxies = None
                     response = requests.get(url, headers=default_headers)
             else:
-                self.logger.warning("Proxy dead. Fetching without proxy")
+                if not noproxy:
+                    self.logger.warning("Proxy dead. Fetching without proxy")
                 response = requests.get(url, headers=default_headers)
             if url.startswith("http:") or response.url.startswith("https:"):
                 return Response(response.content)
@@ -582,7 +633,7 @@ class RefMan:
                            if "href" in x.attrs and x["href"].lower().endswith(".pdf")
                            and re.match(regexp.lower(), x["href"].lower())]
                 if not matches:
-                    return f"{title}"
+                    return f"URL Not found for {title}"
                 elif len(matches) == 1:
                     href = matches[0][0]["href"]
                 else:
